"""
  Utility functions to manipulate structures. based on Bio.PDB data model
"""

import re
import sys
import numpy as np
from numpy import arccos, clip, cos, dot, pi, sin, sqrt
from numpy.linalg import norm

from Bio.PDB.Atom import Atom
from Bio.PDB.NeighborSearch import NeighborSearch
from Bio.PDB.vectors import Vector, rotaxis

UNKNOWN = 0

#chain types
PROTEIN = 1
DNA = 2
RNA = 3
NA = 4
SEQ_THRESHOLD = 0.8
CHAIN_TYPE_LABELS = {PROTEIN:'Protein', DNA:'DNA', RNA:'RNA', UNKNOWN:'Unknown'}

#Model Types
ENSM = 1
BUNIT = 2
MODELS_MAXRMS = 15.0    # Threshold value to detect NMR models (angs)
MODEL_TYPE_LABELS = {ENSM:'Ensembl/NMR', BUNIT:'BioUnit', UNKNOWN:'Unknown'}

#HetAtm Types
MODRES = 1
METAL = 2
ORGANIC = 3
COVORGANIC = 4
WAT = 5

#Geometry
CBINTERNALS = [1.5, 115.5, -123.]
OINTERNALS = [1.229, 120.500, 0.000]
SP3ANGLE = 109.470
SP3DIHS = [60.0, 180.0, 300.0]
HDIS = 1.08


# TODO: consider replace by Bio.PDB equivalent
ONE_LETTER_RESIDUE_CODE = {
    'ALA':'A', 'CYS':'C', 'ASP':'D', 'GLU':'E', 'PHE':'F', 'GLY':'G',
    'HIS':'H', 'HID':'H', 'HIE':'H', 'ILE':'I', 'LYS':'K', 'LEU':'L',
    'MET':'M', 'ASN':'N', 'PRO':'P', 'GLN':'Q', 'ARG':'R', 'SER':'S',
    'THR':'T', 'VAL':'V', 'TRP':'W', 'TYR':'Y'
}

THREE_LETTER_RESIDUE_CODE = {
    'A':'ALA', 'C': 'CYS', 'D':'ASP', 'E':'GLU', 'F':'PHE', 'G':'GLY',
    'H':'HIS', 'I':'ILE', 'K':'LYS', 'L':'LEU', 'M':'MET', 'N':'ASN',
    'P':'PRO', 'Q':'GLN', 'R':'ARG', 'S':'SER',
    'T':'THR', 'V':'VAL', 'W':'TRP', 'Y':'TYR'
}

DNA_RESIDUE_CODE = ['DA', 'DC', 'DG', 'DT']
RNA_RESIDUE_CODE = ['A', 'C', 'G', 'U']
NA_RESIDUE_CODE = DNA_RESIDUE_CODE + RNA_RESIDUE_CODE

# Residue & Atom friendly ids
def residue_id(res, models='auto'):
    """
    Friendly replacement for residue ids like ASN A324/0
    """
    return '{:>3} {}'.format(res.get_resname(), residue_num(res, models))

def residue_num(res, models='auto'):
    """
    Shortcut for getting residue num including chain id, includes model number if any
    """
    if models == 'auto':
        models = len(res.get_parent().get_parent().get_parent()) > 1

    if has_ins_code(res):
        rnum = str(res.get_parent().id) + str(res.id[1]) + res.id[2]
    else:
        rnum = str(res.get_parent().id) + str(res.id[1])
    if models:
        rnum += "/" + str(res.get_parent().get_parent().id + 1)
    return rnum

def atom_id(atm, models='auto'):
    """
    Friendly replacement for atom ids like ASN A324/0.CA
    """
    return '{}.{}'.format(residue_id(atm.get_parent(), models), atm.id)

# Id Checks
def protein_residue_check(res):
    """
    Checks whether is a valid protein residue id, either one or three letter code
    return upper case three-letter code
    """
    res = res.upper()
    rid = ''
    if res in THREE_LETTER_RESIDUE_CODE.keys():
        rid = THREE_LETTER_RESIDUE_CODE[res]
    elif res in ONE_LETTER_RESIDUE_CODE.keys():
        rid = res
    else:
        return False

    return rid

def same_residue(at1, at2):
    """
    Checks whether atoms belong to the same residue
    """
    return at1.get_parent() == at2.get_parent()

def same_model(res1, res2):
    """
    Checks whether residues belong to the same model
    """
    return res1.get_parent().get_parent() == res2.get_parent().get_parent()

def same_chain(res1, res2):
    """
    Checks whether residues belong to the same chain
    """
    return (res1.get_parent() == res2.get_parent()) and same_model(res1, res2)

def seq_consecutive(res1, res2):
    """
    Checks whether residues belong to the same chain and are consecutive in sequences,
    taken from residue number
    """
    rnum1 = res1.id[1]
    rnum2 = res2.id[1]
    return same_chain(res1, res2) and abs(rnum1 - rnum2) == 1

def is_wat(res):
    """
    Shortcut to check for water residues
    """
    return res.id[0].startswith('W')

def is_hetatm(res):
    """
    Shortcut to check for HETATM residues
    """
    return res.id[0].startswith('H_') or res.id[0].startswith('W')

def is_at_in_list(atm, at_list, rname=None):
    """Checks if **at is in **at_list in the context of **rname residue
    """
    if rname is None:
        rname = atm.get_parent().get_resname().replace(' ', '')
    if rname in at_list:
        return atm.id in at_list[rname] or atm.id in at_list['*']
    return atm.id in at_list['*']

def has_ins_code(res):
    """Checks whether residue **r has insertion code
    """
    return res.id[2] != ' '

#===============================================================================
# Guesses

def guess_models_type(struc, threshold=MODELS_MAXRMS):
    """Guesses the type of models according to all atom RMSd of first 2 models
       Considers Ensembl/NMR if rmsd value is less than **threshold
    """
    if len(struc) == 1:
        return 0
    rmsd = calc_RMSd_all_ats(struc[0], struc[1])
    rmsd = round(rmsd, 4)
    if rmsd < threshold:
        return {'type':ENSM, 'rmsd':rmsd}
    return {'type':BUNIT, 'rmsd':rmsd}

def guess_chain_type(chn, thres=SEQ_THRESHOLD):
    """
    Guesses chain type (protein, dna, or rna) from residue composition
    Allow for non-std residues.
    """
    #TODO improve guessing for hybrid chains
    prot = 0.
    dna = 0.
    rna = 0.
    total = 0.
    for res in chn.get_residues():
        total += 1
        if is_wat(res):
            continue
        rname = res.get_resname().replace(' ', '')
        if rname in THREE_LETTER_RESIDUE_CODE.values():
            prot += 1
        elif rname in DNA_RESIDUE_CODE:
            dna += 1
        elif rname in RNA_RESIDUE_CODE:
            rna += 1
    prot = prot / total
    dna = dna / total
    rna = rna / total
    other = 1 - prot - dna - rna
    if prot > thres or prot > dna + rna:
        return PROTEIN
    elif dna > thres or dna > prot + rna:
        return DNA
    elif rna > thres or rna > prot + dna:
        return RNA
    return [prot, dna, rna, other]

#===============================================================================
def check_chiral_residue(res, chiral_data):
    """
        Checks proper chirality of side chain atoms as defined in
        chiral_data
    """
    chi_ok = True
    if res.get_resname() in chiral_data:
        at_ids = chiral_data[res.get_resname()]
        chi_ok = check_chiral(res, 'CB', 'CA', at_ids[0], at_ids[1], -1.)
    return chi_ok

def check_chiral_ca(res):
    """
    Checks proper (L) quirality of CB atom with respect to backbone
    """
    return res.get_resname() == 'GLY' or check_chiral(res, 'N', 'CA', 'C', 'CB')

def check_chiral(res, at1, at2, at3, at4, sign=1.):
    """
    Checks proper chirality of at2 atom.
    at1-at3 define reference plane.
    Position of at4 with respect to the plane is checked.
    Sign (+1,-1) allows to check for a specific enantiomer.
    """
    # check all atoms are present
    at_ok = True
    chi_ok = True
    for atm in [at1, at2, at3, at4]:
        at_ok = at_ok and atm in res
        if not at_ok:
            print('Warning: atom {:3} not found in {}'.format(atm, residue_id(res)))
    if at_ok:
        vec1 = res[at1].coord - res[at2].coord
        vec2 = res[at3].coord - res[at2].coord
        vecp = np.cross(vec1, vec2)
        vec3 = res[at4].coord - res[at2].coord
        chi_ok = sign * (_calc_v_angle(vecp, vec3) - 90.) < 0.
    return chi_ok

def check_all_at_in_r(res, at_list):
    """ Check whether all residue atoms are present. """
    miss_at = {}
    for group in ['backbone', 'side']:
        miss_at[group] = []
        for at_id in at_list[group]:
            if not at_id in res:
                miss_at[group].append(at_id)
    if miss_at['backbone'] + miss_at['side']:
        return miss_at
    return {}

def check_r_list_clashes(r_list, rr_list, clash_dist, atom_lists, join_models=True):
    """ Check clashes generated by a list of residues """
    clash_list = {'severe':{}}
    for cls in atom_lists:
        clash_list[cls] = {}
    for r_pair in rr_list:
        [res1, res2] = r_pair[0:2]

        if (res1 in r_list or res2 in r_list) and not is_wat(res1) and not is_wat(res2):
            c_list = check_rr_clashes(res1, res2, clash_dist, atom_lists, join_models)
            rkey = residue_id(res1) + '-' + residue_id(res2)
            for cls in c_list:
                if c_list[cls]:
                    clash_list[cls][rkey] = c_list[cls]
    return clash_list

def check_rr_clashes(res1, res2, clash_dist, atom_lists, join_models=True):
    """ Check all clashes between two residues """
    clash_list = {}
    min_dist2 = {}
    clash_dist2 = {}
    ats_list1 = {}
    ats_list2 = {}
    for cls in atom_lists:
        clash_list[cls] = []
        min_dist2[cls] = 99999.
        clash_dist2[cls] = clash_dist[cls]**2
        ats_list1[cls] = set()
        ats_list2[cls] = set()
    for atm in res1.get_atoms():
        for cls in atom_lists:
            if is_at_in_list(atm, atom_lists[cls], res1.get_resname()):
                ats_list1[cls].add(atm.id)
    for atm in res2.get_atoms():
        for cls in atom_lists:
            if is_at_in_list(atm, atom_lists[cls], res2.get_resname()):
                ats_list2[cls].add(atm.id)

    if res1 != res2 and not seq_consecutive(res1, res2) \
                and (join_models or same_model(res1, res2)):
        for at_pair in get_all_rr_distances(res1, res2):
            [at1, at2, dist2] = at_pair
            if 'severe' in atom_lists and dist2 < clash_dist2['severe']:
                if dist2 < min_dist2:
                    clash_list['severe'] = at_pair
                    min_dist2['severe'] = dist2
            else:
                for cls in atom_lists:
                    if cls == 'apolar':
                        #Only one of the atoms should be apolar
                        if not at1.id in ats_list1[cls] and \
                                not at2.id in ats_list2[cls]:
                            continue
                        #Remove n->n+2 backbone clashes. TODO Improve
                        if abs(res1.index - res2.index) <= 2:
                            continue
                        #Remove Ca2+ looking like backbone CA's
                        if at1.id == 'CA' and is_hetatm(res1) or \
                                at2.id == 'CA' and is_hetatm(res2):
                            continue
                    else:
                        # Both atoms should be of the same kind
                        if not at1.id in ats_list1[cls] or \
                                not at2.id in ats_list2[cls]:
                            continue
                    if dist2 < clash_dist2[cls]:
                        if dist2 < min_dist2[cls]:
                            clash_list[cls] = at_pair
                            min_dist2[cls] = dist2
    return clash_list

#===============================================================================
def get_altloc_residues(struc):
    """ Gets list of residue  with atoms with alternative location labels
    """
    res_list = {}
    for res in struc.get_residues():
        for atm in res.get_atoms():
            if atm.get_altloc() != ' ':
                if res not in res_list:
                    res_list[res] = []
                res_list[res].append(atm)
    return res_list

def get_metal_atoms(struc, metal_ats):
    """
    Gets list of metal atoms

    """
    met_list = []
    for atm in struc.get_atoms():
        if not re.match('H_', atm.get_parent().id[0]):
            continue
        if atm.id in metal_ats:
            met_list.append(atm)
    return met_list

def get_ligands(struc, incl_water=False):
    """
    Gets lists of ligands, water molecules can be excluded
    """
    lig_list = []
    for res in struc.get_residues():
        if is_hetatm(res) and (incl_water or not is_wat(res)):
            lig_list.append(res)
    return lig_list

def get_residues_with_H(struc):
    """ Get residues containing Hydrogen atoms
    """
    resh_list = []
    for res in struc.get_residues():
        has_h = 0
        for atm in res.get_atoms():
            if atm.element == 'H':
                has_h += 1
        if has_h:
            resh_list.append({'r':res, 'n_h':has_h})
    return resh_list


def get_backbone_links(struc, backbone_atoms, covlnk, join_models=True):
    """ Get links making the main chain """
    #TODO differenciate Protein and NA
    cov_links = []
    for mod in struc:
        bckats = []
        for atm in struc[mod.id].get_atoms():
            if atm.id in backbone_atoms:
                if atm.disordered_flag:
                    bckats.append(atm.selected_child)
                else:
                    bckats.append(atm)
        if bckats:
            nbsearch = NeighborSearch(bckats)
            for at1, at2 in nbsearch.search_all(covlnk):
                if not same_residue(at1, at2) \
                        and (join_models or same_model(at1.get_parent(), at2.get_parent())):
                    cov_links.append(sorted([at1, at2], key=lambda x: x.serial_number))
        else:
            print("Warning: No backbone atoms defined")

    return cov_links

# Residue manipulation =======================================================
def remove_H_from_r(res, verbose=False):
    """
    Removes Hydrogen atoms from given residue
    """
    h_list = []
    for atm in res.get_atoms():
        if atm.element == 'H':
            h_list.append(atm.id)
    for at_id in h_list:
        if verbose:
            print("  Deleting atom " + at_id)
        res.detach_child(at_id)

def remove_residue(res):
    """
    Removes residue completely
    """
    res.get_parent().detach_child(res.id)

def swap_atoms(at1, at2):
    """
    Swaps names for two given atoms. Useful to fix labelling issues
    """
    at1_id = at1.id
    at1_full_id = at1.full_id
    at1_element = at1.element
    at1_name = at1.name
    at1_fullname = at1.fullname

    at1.id = at2.id
    at1.full_id = at2.full_id
    at1.element = at2.element
    at1.name = at2.name
    at1.fullname = at2.fullname

    at2.id = at1_id
    at2.full_id = at1_full_id
    at2.element = at1_element
    at2.name = at1_name
    at2.fullname = at1_fullname

#def invert_chirality(res, at1, at2, at3, at4):
#    """Inverts chirality of at2 by rotating at4, and the associated end chain atoms
    #"""
#    #TODO
#
#def invert_chiral_ca(res):
#    """
#    Inverts CA Chirality.
    #"""
#    #TODO

# Atom management ==============================================================
def rename_atom(res, old_at, new_at):
    atm = res[old_at]
    res.detach_child(atm.id)
    atm.id = new_at
    atm.full_id = new_at
    atm.element = new_at[0:1]
    atm.fullname = ' ' + new_at
    res.add(atm)

def delete_atom(res, at_id):
    res.detach_child(at_id)

def add_hydrogens_backbone(res, res_1):
    """ Add hydrogen atoms to the backbone"""

    # only proteins

    if not protein_residue_check(res.get_resname()):
        return "Warning: Residue not valid in this context "

    error_msg = "Warning: not enough atoms to build backbone hydrogen atoms on"

    if 'N' not in res or 'CA' not in res:
        return error_msg

    if res_1 is None:
        # Nterminal TODO  Neutral NTerm
        if res.get_resname() == 'PRO':
            if 'CD' not in res:
                return error_msg

            add_new_atom_to_residue(
                res,
                'H',
                build_coords_SP2(HDIS, res['N'], res['CA'], res['CD'])
            )
        else:
            if 'C' not in res:
                return error_msg

            crs = build_coords_3xSP3(HDIS, res['N'], res['CA'], res['C'])
            add_new_atom_to_residue(res, 'H1', crs[0])
            add_new_atom_to_residue(res, 'H2', crs[1])
            add_new_atom_to_residue(res, 'H3', crs[2])

    elif res.get_resname() != 'PRO':
        if 'C' not in res_1:
            return error_msg

        add_new_atom_to_residue(
            res,
            'H',
            build_coords_SP2(HDIS, res['N'], res['CA'], res_1['C'])
        )

    if res.get_resname() == 'GLY':
        if 'C' not in res:
            return error_msg

        crs = build_coords_2xSP3(HDIS, res['CA'], res['N'], res['C'])
        add_new_atom_to_residue(res, 'HA2', crs[0])
        add_new_atom_to_residue(res, 'HA3', crs[1])

    else:
        if 'C' not in res or 'CB' not in res:
            return error_msg

        add_new_atom_to_residue(
            res,
            'HA',
            build_coords_1xSP3(HDIS, res['CA'], res['N'], res['C'], res['CB'])
        )

    return False



def add_hydrogens_side(res, res_library, opt, rules):
    """ Add hydrogens to side chains"""
    if 'N' not in res or 'CA' not in res or 'C' not in res:
        return "Warning: not enough atoms to build side chain hydrogen atoms on"

    for key_rule in rules.keys():
        rule = rules[key_rule]
        if rule['mode'] == 'B2':
            crs = build_coords_2xSP3(
                rule['dist'],
                res[key_rule],
                res[rule['ref_ats'][0]],
                res[rule['ref_ats'][1]]
            )
            add_new_atom_to_residue(res, rule['ats'][0], crs[0])
            add_new_atom_to_residue(res, rule['ats'][1], crs[1])

        elif rule['mode'] == "B1":
            crs = build_coords_1xSP3(
                rule['dist'],
                res[key_rule],
                res[rule['ref_ats'][0]],
                res[rule['ref_ats'][1]],
                res[rule['ref_ats'][2]]
            )
            add_new_atom_to_residue(res, rule['ats'][0], crs)

        elif rule['mode'] == 'S2':
            crs = build_coords_SP2(
                rule['dist'],
                res[key_rule],
                res[rule['ref_ats'][0]],
                res[rule['ref_ats'][1]],
            )
            add_new_atom_to_residue(res, rule['ats'][0], crs)

        elif rule['mode'] == 'L':
            for at_id in rule['ats']:
                crs = build_coords_from_lib(res, res_library, opt, at_id)
                add_new_atom_to_residue(res, at_id, crs)
    return False

def build_atom(res, at_id, res_lib, new_res_id):
    if at_id == 'CB':
        coords = build_coords_CB(res)
    else:
        coords = build_coords_from_lib(res, res_lib, new_res_id, at_id)
    add_new_atom_to_residue(res, at_id, coords)


def add_new_atom_to_residue(res, at_id, coords):
    res.add(Atom(at_id, coords, 99.0, 1.0, ' ', ' ' + at_id + ' ', 0, at_id[0:1]))


def build_coords_from_lib(res, res_lib, new_res, at_id):
    """
     Calculates cartesian coordinates for a new atom from internal coordinates definition.

    """
    atom_def = res_lib.get_atom_def(new_res, at_id)
    if atom_def is None:
        print("#ERROR: Unknown target atom")
        sys.exit(1)

    return build_coords_from_ats_internal(
        res[atom_def.link_ats[0]],
        res[atom_def.link_ats[1]],
        res[atom_def.link_ats[2]],
        atom_def.geom
    )

def build_coords_CB(res): # Get CB from Backbone
    """
     Calculates cartesian coordinates for a new CB atom from backbone.

    """
    return build_coords_from_ats_internal(res['CA'], res['N'], res['C'], CBINTERNALS)

def build_coords_O(res): # Get O from Backbone
    """
     Calculates cartesian coordinates for a new O atom from backbone.

    """
    return build_coords_from_ats_internal(res['C'], res['CA'], res['N'], OINTERNALS)

def build_coords_3xSP3(dst, at0, at1, at2):
    """
        Generates coordinates for 3 SP3 atoms
        **dst** bond distance
        **at0**  central atom
        **at1** atom to define bond angles
        **at2** atom to define dihedrals
    """
    #TODO try a pure geometrical generation to avoid at2
    crs = []
    for i in range(0, 3):
        crs.append(build_coords_from_ats_internal(at0, at1, at2, [dst, SP3ANGLE, SP3DIHS[i]]))
    return crs

def build_coords_2xSP3(dst, at0, at1, at2):
    """
        Generates coordinates for two SP3 bonds given the other two
        **dst** Bond distance
        **at0** Central atom
        **at1** atom with existing bond
        **at2** atom with existing bond
    """
    cr0 = Vector(at0.get_coord())
    cr1 = Vector(at1.get_coord())
    cr2 = Vector(at2.get_coord())
    axe = cr0 - cr1
    mat = rotaxis(120.*pi/180., axe)
    bond = cr2 - cr0
    bond.normalize()
    bond._ar = bond._ar * dst
    cr3 = cr0 + bond.left_multiply(mat)
    cr4 = cr0 + bond.left_multiply(mat).left_multiply(mat)
    crs = []
    crs.append(cr3._ar)
    crs.append(cr4._ar)
    return crs

def build_coords_1xSP3(dst, at0, at1, at2, at3):
    """
      Calculated cartesian coordinates to complete a SP3 group
    """
    cr0 = at0.get_coord()
    cr1 = at1.get_coord()
    cr2 = at2.get_coord()
    cr3 = at3.get_coord()
    avg = cr1 + cr2
    avg = avg + cr3
    avg /= 3.
    avec = cr0 - avg
    avec /= norm(avec)
    avec *= dst
    return cr0 + avec

def build_coords_SP2(dst, at0, at1, at2):
    """
      Calculates cartesian coordinaties to complete a SP2 group
    """
    cr0 = at0.get_coord()
    cr1 = at1.get_coord()
    cr2 = at2.get_coord()

    avg = cr1 + cr2
    avg /= 2.
    avec = cr0 - avg
    avec /= norm(avec)
    avec *= dst
    return cr0 + avec


def build_coords_from_ats_internal(at1, at2, at3, geom):
    return build_coords_from_internal(
        at1.get_coord(),
        at2.get_coord(),
        at3.get_coord(),
        geom)

def build_coords_from_internal(at1c, at2c, at3c, geom):
    """
     Calculates cartesian coordinates for a new atom from internal coordinates.

    """
    dst = geom[0]
    ang = geom[1] * pi / 180.
    tor = geom[2] * pi / 180.0

    vec1 = at1c - at2c
    vec2 = at1c - at3c

    vcr12 = np.cross(vec1, vec2)
    vcr112 = np.cross(vec1, vcr12)

    vcr12 /= norm(vcr12)
    vcr112 /= norm(vcr112)

    vcr12 *= -sin(tor)
    vcr112 *= cos(tor)

    vec3 = vcr12 + vcr112
    vec3 /= norm(vec3)
    vec3 *= dst * sin(ang)

    vec1 /= norm(vec1)
    vec1 *= dst * cos(ang)

    return at1c + vec3 - vec1

# Metrics =============================================================
def calc_at_dist(at1, at2):
    """
    Calculates distance between two atoms
    """
    return np.sqrt(calc_at_sq_dist(at1, at2))

def calc_at_sq_dist(at1, at2):
    """
    Calculates distance between two atoms
    """
    vec = at1.coord - at2.coord
    return np.dot(vec, vec)

def calc_bond_angle(at1, at2, at3):
    """
    Calculates angle among three atoms at1-at2-at3
    """
    vec1 = at1.coord - at2.coord
    vec2 = at3.coord - at2.coord
    return _calc_v_angle(vec1, vec2)

def calc_bond_dihedral(at1, at2, at3, at4):
    """
    Calculates dihedral angles at1-at2-at3-at4
    """
    abv = at1.coord - at2.coord
    cbv = at3.coord - at2.coord
    dbv = at4.coord - at3.coord
    uvec = np.cross(abv, cbv)
    vvec = np.cross(dbv, cbv)
    wvec = np.cross(uvec, vvec)
    angle_uv = _calc_v_angle(uvec, vvec)
    if norm(wvec) == 0.:
        angle_cbw = 0.
    else:
        angle_cbw = _calc_v_angle(cbv, wvec)
    try:
        if angle_cbw > 0.001:
            angle_uv = -angle_uv
    except ZeroDivisionError:
        pass
    return angle_uv

def get_all_at2at_distances(struc, at_ids='all', d_cutoff=0., join_models=False):
    """
    Gets a list of all at-at distances below a cutoff, at ids can be limited
    """
    if not isinstance(at_ids, list):
        at_ids = at_ids.split(',')

    dist_mat = []
    at_list = []
    d_cut2 = d_cutoff**2
    for atm in struc.get_atoms():
        if atm.id in at_ids or at_ids == ['all']:
            at_list.append(atm)
    for i in range(len(at_list)-1):
        for j in range(i + 1, len(at_list)):
            if join_models or same_model(at_list[i].get_parent(), at_list[j].get_parent()):
                dist2 = calc_at_sq_dist(at_list[i], at_list[j])
                if d_cutoff > 0. and dist2 < d_cut2:
                    dist_mat.append([at_list[i], at_list[j], dist2])
    return dist_mat


def get_all_r2r_distances(struc, r_ids='all', d_cutoff=0., join_models=False):
    # Uses distances between the first atom of each residue as r-r distance
    if not isinstance(r_ids, list):
        r_ids = r_ids.split(',')
    dist_mat = []
    check_ats = {}
    for md_id in range(len(struc)):
        check_ats[md_id] = []
        for res in struc[md_id].get_residues():
            if res.resname in r_ids or r_ids == ['all']:
                if join_models:
                    check_ats[0].append(res.child_list[0])
                else:
                    check_ats[md_id].append(res.child_list[0])
    for md_id in range(len(struc)):
        if check_ats[md_id]:
            dist_mat += _get_contacts(check_ats[md_id], d_cutoff)
    return dist_mat

def _get_contacts(ats_list, d_cutoff):
    contact_list = []
    nbsearch = NeighborSearch(ats_list)
    for at1, at2 in nbsearch.search_all(d_cutoff):
        contact_list.append([at1.get_parent(), at2.get_parent(), at1 - at2])
    return contact_list

def calc_RMSd_ats(ats1, ats2):
    if len(ats1) != len(ats2):
        print(
            "Warning: atom lists of different length when calculating RMSd ({}, {})".format(
                len(ats1), len(ats2)
            )
        )
    rmsd = 0
    i = 0
    while i < len(ats1) and i < len(ats2):
        dist2 = calc_at_sq_dist(ats1[i], ats2[i])
        rmsd = rmsd + dist2
        i = i + 1

    return sqrt(rmsd / i)

def calc_RMSd_all_ats(st1, st2):
    ats1 = []
    ats2 = []

    for atm in st1.get_atoms():
        ats1.append(atm)
    for atm in st2.get_atoms():
        ats2.append(atm)

    return calc_RMSd_ats(ats1, ats2)

def get_all_rr_distances(res1, res2, with_h=False):
    dist_mat = []
    for at1 in res1.get_atoms():
        if at1.element == 'H' and not with_h:
            continue
        for at2 in res2.get_atoms():
            if at2.element == 'H' and not with_h:
                continue
            dist2 = calc_at_sq_dist(at1, at2)
            dist_mat.append([at1, at2, dist2])
    return dist_mat
#===============================================================================
def _calc_v_angle(vec1, vec2, deg=True):
    angle = arccos(clip(dot(vec1, vec2)/norm(vec1)/norm(vec2), -1., 1.))
    if deg:
        angle *= 180./pi
    return angle



"""
  Module to manage mutations
"""

import re
import sys

import biobb_structure_manager.model_utils as mu

ADD = 'Add'
DEL = 'Del'
MOV = 'Mov'

class MutationManager():
    """ Class to manage list of mutations. """
    def __init__(self, id_list):
        self.mutation_list = []

        if 'file:' in id_list:
            #Load from file
            id_list = id_list.replace('file:', '')
            print('Reading mutation list from file {}'.format(id_list))
            self.id_list = []
            for line in open(id_list, 'r'):
                self.id_list.append(line.replace('\n', '').replace('\r', ''))

        else:
            self.id_list = id_list.replace(' ', '').split(',')

        #convert to list ids to MutationSet instances and make the list
        self.mutation_list = list(map(MutationSet, self.id_list))
    
    def prepare_mutations(self, struc, stop_on_error=True):
        """  Check mutation_list and unroll chains/models."""
        for mut in self.mutation_list:
            mut.prepare(struc, stop_on_error)

    def apply_mutations(self, struc, mutation_map, residue_lib, remove_h='mut'):
        """ Perform the requestet mutations."""
        mutated_res = []
        for mut in self.mutation_list:
            mutated_res += mut.apply(struc, mutation_map, residue_lib, remove_h)
        return mutated_res

    def __str__(self):
        return ','.join(self.id_list)

#===============================================================================
class MutationSet():
    """ Class to manage instances of a mutation. """
    def __init__(self, mut_id):
        if ':' not in mut_id:
            mut_id = '*:' + mut_id

        self.id = mut_id.upper()

        [self.chain, mut] = mut_id.split(':')

        mut_comps = re.match('([A-z]*)([0-9]*)([A-z]*)', mut)

        self.old_id = mu.protein_residue_check(mut_comps.group(1))
        self.new_id = mu.protein_residue_check(mut_comps.group(3))
        self.res_num = mut_comps.group(2)

        self.id = ''.join([self.chain, ":", self.old_id, self.res_num, self.new_id])
        self.mutations = []

    def prepare(self, struc, stop_on_error=True):
        """ Check which mutations are possible. """
        mut_ok = 0
        for model in struc.get_models():
            if self.chain == '*':
                chains = []
                for chn in model.get_list():
                    chains.append(chn.get_id())
            else:
                chains = [self.chain]

            for chn in chains:
                if int(self.res_num) in model[chn]:
                    res = model[chn][int(self.res_num)]
                    if res.get_resname() == self.old_id:
                        self.mutations.append({
                            'model':model.get_id(),
                            'chain':chn,
                            'residue':res.get_id(),
                            'new_id':self.new_id
                        })
                        mut_ok += 1
                    else:
                        print(
                            '#WARNING: Unknown residue {}:{}{}'.format(
                                chn, self.old_id, self.res_num
                            )
                        )
                else:
                    print(
                        '#WARNING: Unknown residue {}:{}{}', format(
                            chn, self.old_id, self.res_num
                        )
                    )

        if not mut_ok and stop_on_error:
            print(
                '#ERROR: no mutations available for {}'.format(self.id), 
                file=sys.stderr
            )
            sys.exit(1)

    def apply(self, struc, mut_map, res_lib, remove_h):
        """ Perform the individual mutations on the set. """
        mutated_res = []
        for mut in self.mutations:
            res = struc[mut['model']][mut['chain']][mut['residue']]
            rname = res.get_resname().replace(' ', '')
            # Deleting H
            if remove_h == 'mut':
                mu.remove_H_from_r(res, verbose=True)
            # checking side chain
            side_atoms = []
            bck_atoms = []
            for atm in res.get_atoms():
                if atm.id in mut_map[rname]['side_atoms']:
                    side_atoms.append(atm.id)
                else:
                    bck_atoms.append(atm.id)
            for at_id in ['N', 'CA', 'C']:
                if at_id not in bck_atoms:
                    print(
                        '#ERROR: Backbone atoms missing for {}, aborting'.format(
                            mu.residue_id(res)
                        ), file=sys.stderr
                    )
                    sys.exit(1)
            missing_ats = []
            for at_id in mut_map[rname]['side_atoms']:
                if at_id not in side_atoms:
                    missing_ats.append(at_id)
            print("Replacing " + mu.residue_id(res) + " into " + self.new_id)
            in_rules = []
            extra_adds = []
            # Renaming ats
            for rule in mut_map[rname][self.new_id][MOV]:
                [old_at, new_at] = rule.split("-")
                print('  Renaming {} to {}'.format(old_at, new_at))
                if old_at in side_atoms:
                    mu.rename_atom(res, old_at, new_at)
                else:
                    print(
                        '#WARNING: atom {} missing in {}'.format(
                            old_at, mu.residue_id(res)
                        )
                    )
                    extra_adds.append(new_at)
                in_rules.append(old_at)
            # Deleting atoms
            for at_id in mut_map[rname][self.new_id][DEL]:
                print('  Deleting {}'.format(at_id))
                if at_id in side_atoms:
                    mu.delete_atom(res, at_id)
                else:
                    print(
                        '#WARNING: atom {} already missing in {}'.format(
                            at_id, mu.residue_id(res)
                        )
                    )
                in_rules.append(at_id)
            # Adding atoms (new_id required as r.resname is still the original)
            # Adding missing atoms that keep name
            for at_id in mut_map[rname]['side_atoms']:
                if at_id not in in_rules and at_id in missing_ats:
                    print('  Adding missing atom {}'.format(at_id))
                    mu.build_atom(res, at_id, res_lib, self.new_id)
            for at_id in extra_adds:
                print('  Adding new atom {}'.format(at_id))
                mu.build_atom(res, at_id, res_lib, self.new_id)
            for at_id in mut_map[rname][self.new_id][ADD]:
                print('  Adding new atom {}'.format(at_id))
                mu.build_atom(res, at_id, res_lib, self.new_id)
            #Renaming residue
            res.resname = self.new_id
            mutated_res.append(res)
        print("")
        return mutated_res

    def __str__(self):
        return self.id



"""
  Module to load data library json file
"""

import json
import sys

class DataLibManager():
    """ Manages projects' global data file
    """

    def __init__(self, file_path):
        try:
            data_file_h = open(file_path, "r")
            json_map = json.load(data_file_h)
            self.residue_codes = json_map['data_library']['residue_codes']
            self.atom_data = json_map['data_library']['atom_data']
            self.residue_data = json_map['data_library']['residue_data']
            self.distances = json_map['data_library']['distances']
            self.ion_res = json_map['data_library']['addH_check_residues']
            self.std_ion = json_map['data_library']['addH_std_ion']

        except IOError:
            print("ERROR: unable to open data library " + file_path, file=sys.stderr)
            sys.exit(2)

    def get_valid_codes(self, mol_type):
        """ Obtain valid residue codes """
        if mol_type == 'na':
            codes = self.residue_codes['dna'] + self.residue_codes['rna']
        else:
            codes = self.residue_codes[mol_type]
        return codes

    def get_all_atom_lists(self):
        """ Obtains lists of atoms per protein residue. """
        ats_lists = {}
        for code in self.residue_codes['protein']:
            ats_lists[code] = {
                'backbone': self.residue_data['*']['bck_atoms'],
                'side': self.residue_data[code]['side_atoms']
            }
        return ats_lists

    def get_metal_atoms(self):
        """ Gets metal atom names. """
        return self.atom_data['metal_atoms']

    def get_atom_feature_list(self, feature):
        """ Gets a residue list with a specific section of data . """
        f_list = {}
        for rcode in self.residue_data:
            if feature in self.residue_data[rcode]:
                f_list[rcode] = self.residue_data[rcode][feature]
        if '*' not in f_list:
            f_list['*'] = []
        return f_list

    def get_atom_lists(self, contact_types):
        """ Gets a list of atoms organized per contact types. """
        atom_lists = {}
        for cls in contact_types:
            if cls != 'severe':
                atom_lists[cls] = self.get_atom_feature_list(cls + '_atoms')
        return atom_lists

    def get_mutation_rules(self, aa_in, aa_out, rule_group):
        """ Gets mutation rules for a specific mutation, and rule type """
        return self.residue_data[aa_in]['mutation_rules'][aa_out][rule_group]

    def get_distances(self, distid):
        """ Get distance constant. """
        return self.distances[distid]

    def get_amide_data(self):
        """ Gets data related to amide residues """
        alist = []
        rlist = {}
        for rcode in self.residue_data:
            if 'amide_atoms' in self.residue_data[rcode]:
                alist += self.residue_data[rcode]['amide_atoms']
                rlist[rcode] = self.residue_data[rcode]['amide_atoms']
        return [rlist, alist]

    def get_chiral_data(self):
        """ Gets data related to chiral residues. """
        rlist = {}
        for rcode in self.residue_data:
            if 'chiral_atoms' in self.residue_data[rcode]:
                rlist[rcode] = self.residue_data[rcode]['chiral_atoms']
        return rlist

    def get_mutation_map(self):
        """ Gets the complete map of mutation rules per residue."""
        mut_rules = {}
        for rcode in self.residue_data:
            if 'mutation_rules' in self.residue_data[rcode]:
                mut_rules[rcode] = self.residue_data[rcode]['mutation_rules']
                mut_rules[rcode]['side_atoms'] = self.residue_data[rcode]['side_atoms']
        return mut_rules

    def get_hydrogen_atoms(self):
        """ Gets list of hydrogen atoms per residue. """
        h_atom_list = {}
        for rcode in self.residue_data:
            if 'hydrogen_atoms' in self.residue_data[rcode]:
                h_atom_list[rcode] = self.residue_data[rcode]['hydrogen_atoms']
        return h_atom_list

    def get_ion_data(self):
        """ Gets residues data related to ionization states. """
        return self.std_ion

    def get_add_h_rules(self):
        """ Gets rules for adding Hydrogen atoms to residues. """
        rules_list = {}
        for rcode in self.residue_data:
            if 'addH_rules' in self.residue_data[rcode]:
                rules_list[rcode] = self.residue_data[rcode]['addH_rules']
        return rules_list



"""Module to manage structure, based on BioPython Bio.PDB
"""
import warnings
import os
from Bio import BiopythonWarning
from Bio.PDB.MMCIF2Dict import MMCIF2Dict
from Bio.PDB.MMCIFParser import MMCIFParser
from Bio.PDB.PDBIO import PDBIO
from Bio.PDB.PDBParser import PDBParser
from Bio.PDB.parse_pdb_header import parse_pdb_header
from biobb_structure_manager.mmb_server import MMBPDBList
from biobb_structure_manager.mutation_manager import MutationManager
from biobb_structure_manager.data_lib_manager import DataLibManager
from biobb_structure_manager.residue_lib_manager import ResidueLib

import biobb_structure_manager.model_utils as mu

CISTHRES = 20  #TODO check vaules with pdb checking
TRANSTHRES = 160

ALL_CONTACT_TYPES = [
    'severe',
    'apolar',
    'polar_acceptor',
    'polar_donor',
    'positive',
    'negative'
]
AMIDE_CONTACT_TYPES = [
    'polar_acceptor',
    'polar_donor',
]

class StructureManager():
    """Main Class wrapping Bio.PDB structure object
    """

    def __init__(
            self,
            input_pdb_path,
            data_library_path,
            res_library_path,
            pdb_server='ftp://ftp.wwpdb.org',
            cache_dir='tmpPDB',
            file_format='mmCif'):
        """Class constructor. Sets an empty object and loads a structure
        according to parameters

        Args:
            **input_pdb_path** (str): path to input structure either in pdb or
            mmCIF format. Format is taken from file extension.
            Alternatively **pdb:pdbId** fetches the mmCIF file from RCSB

            **pdb_server** (str) :  **default** for Bio.PDB defaults (RCSB),
                                    **mmb** for MMB PDB API

        Object structure:
            {
                "input_format" (str): Format of input file pdb|cif
                "models_type" (int): Guessed model type when num. models > 1 NMR|Bunit
                "num_ats" (int): Total Number of atoms
                "nmodels" (int): Number of Models
                "chain_ids" (List): Chain composition as
                    [chain_id:{}]
                "modified" (Boolean): Flag to indicated that structure has been modified
                "all_residues" (list): List of pointer to Bio.PDB.Residue objects, ordered
                    acording to input file
                "num_res" (int): Number of residues
                "res_insc" (int): Number of residues with insertion codes
                "res_hetats" (int): Number of residues flagged as HETATM
                "res_ligands" (int): Number of non water residues flagged as HETATM
                "num_wat" (int): Number of water residues
                "ca_only" (boolead): Flag to indicate a possible CA-only structure
                "modified_residue_list" []: List of residues being connected
                                            HETATM as PDB.Bio.Residue
                "backbone_links" []: List of found canonical backbone links
                                    as [at1, at2] tuples, according to a distance criterium

        """

        self.chain_ids = {}
        self.sorted_chain_ids = []

        self.backbone_links = []
        self.modified_residue_list = []

        self.hetatm = {}
        self.num_res = 0
        self.num_ats = 0
        self.res_hetats = 0
        self.num_wat = 0
        self.res_insc = 0
        self.res_ligands = 0
        self.ca_only = False

        self.meta = {}

        self.all_residues = []
        self.next_residue = {}
        self.prev_residue = {}

        self.modified = False
        self.biounit = False

        self.file_format = file_format

        self.data_library = DataLibManager(data_library_path)
        self.res_library = ResidueLib(res_library_path)

        self.input_format = self._load_structure_file(input_pdb_path, cache_dir, pdb_server, file_format)

        #checking models type according to RMS among models
        self.nmodels = len(self.st)
        if self.nmodels > 1:
            self.models_type = mu.guess_models_type(self.st)
        else:
            self.models_type = 0

        # Calc internal data
        self.update_internals()

    def _load_structure_file(self, input_pdb_path, cache_dir, pdb_server, file_format):
        if "pdb:"in input_pdb_path:
            # MMBPDBList child defaults to Bio.PDB.PDBList if MMB server is not selected
            pdbl = MMBPDBList(pdb=cache_dir, server=pdb_server)
            if '.' in input_pdb_path:
                [pdbid, biounit] = input_pdb_path.split('.')
                input_pdb_path = pdbid[4:].upper()
                if pdb_server != 'mmb':
                    raise WrongServerError
                real_pdb_path = pdbl.retrieve_pdb_file(
                    input_pdb_path, file_format='pdb', biounit=biounit
                )
                self.biounit = biounit
            else:
                input_pdb_path = input_pdb_path[4:].upper()
                real_pdb_path = pdbl.retrieve_pdb_file(
                    input_pdb_path, file_format=self.file_format
                )
                if file_format == 'pdb':
                    # change file name to id.pdb
                    os.rename(real_pdb_path, input_pdb_path + ".pdb")
                    real_pdb_path = input_pdb_path + ".pdb"
        else:
            real_pdb_path = input_pdb_path

        if '.pdb' in real_pdb_path:
            parser = PDBParser(PERMISSIVE=1)
            input_format = 'pdb'
        elif '.cif' in real_pdb_path:
            parser = MMCIFParser()
            input_format = 'cif'
        else:
            raise UnknownFileTypeError(input_pdb_path)
        warnings.simplefilter('ignore', BiopythonWarning)
        self.st = parser.get_structure('st', real_pdb_path)
        if input_format == 'pdb':
            self.headers = parse_pdb_header(real_pdb_path)
        else:
            self.headers = MMCIF2Dict(real_pdb_path)
            
        return input_format

    def update_internals(self):
        """ Update internal data when structure is modified """
        # Add .index field for correlative, unique numbering of residues
        self.residue_renumbering()
        #Atom renumbering for mmCIF, PDB uses atom number in file
        self.atom_renumbering()
        self.set_chain_ids()
        self.calc_stats()
        self.guess_hetatm()
        #Pre_calc rr distances with separated models
        self.rr_dist = mu.get_all_r2r_distances(
            self.st,
            'all',
            self.data_library.get_distances('R_R_CUTOFF'),
            join_models=False
        )
        #Precalc backbone . TODO Nucleic Acids
        self.check_backbone_connect(
            ['N', 'C'],
            self.data_library.get_distances('COVLNK')
        )

    def residue_renumbering(self):
        """Sets the Bio.PDB.Residue.index attribute to residues for a unique,
        consecutive, residue number, and generates the corresponding list
        in **all_residues**
        """
        i = 1
        self.all_residues = []
        for res in self.st.get_residues():
            res.index = i
            if type(res).__name__ == 'DisorderedResidue':
                for ch_r in res.child_dict:
                    res.child_dict[ch_r].index = i
            self.all_residues.append(res)
            i += 1

    def atom_renumbering(self):
        """ Sets  Bio.PDB.Atom.serial_number for all atoms in the structure,
            overriding original if any.
        """
        i = 1
        for atm in self.st.get_atoms():
            atm.serial_number = i
            if hasattr(atm, 'selected_child'):
                atm.selected_child.serial_number = i
            i += 1

    def guess_hetatm(self):
        """ Guesses HETATM type as modified res, metal, wat, organic
        """
        self.hetatm = {}
        for typ  in [mu.UNKNOWN, mu.MODRES, mu.METAL, mu.ORGANIC, mu.COVORGANIC, mu.WAT]:
            self.hetatm[typ] = []
        for res in self.get_structure().get_residues():
            if not mu.is_hetatm(res):
                continue
            if mu.is_wat(res):
                self.hetatm[mu.WAT].append(res)
            elif len(res) == 1:
                self.hetatm[mu.METAL].append(res)
            elif 'N' in res or 'C' in res:
                # modified aminoacid candidate, TODO check connectivity with n-1 or n+1
                self.hetatm[mu.MODRES].append(res)
                #TODO check modified nucleotides
            else:
                self.hetatm[mu.ORGANIC].append(res)

    def calc_stats(self):
        """ Calculates general statistics about the structure, and guesses whether
            it is a CA-only structure
        """
        self.num_res = 0
        self.num_ats = 0
        self.res_hetats = 0
        self.num_wat = 0
        self.res_insc = 0
        for res in self.st.get_residues():
            self.num_res += 1
            if mu.is_wat(res):
                self.num_wat += 1
            if mu.is_hetatm(res):
                self.res_hetats += 1
            if mu.has_ins_code(res):
                self.res_insc += 1
            self.num_ats += len(res.get_list())
        self.res_ligands = self.res_hetats - self.num_wat
        # Detecting whether it is a CA-only structure
        # num_ats should be much larger than num_res
        # waters removed
        # Taking polyGly as a lower limit
        self.ca_only = self.num_ats - self.num_wat < (self.num_res - self.num_wat)*4

    def get_ins_codes(self):
        """Makes a list with residues having insertion codes"""

        ins_codes_list = []
        for res in self.st.get_residues():
            if mu.has_ins_code(res):
                ins_codes_list.append(res)
        return ins_codes_list

    def get_metal_atoms(self):
        """ Makes a list of possible metal atoms"""
        return mu.get_metal_atoms(self.st, self.data_library.get_metal_atoms())

    def get_SS_bonds(self):
        """ Returns possible SS Bonds by distance"""
        return mu.get_all_at2at_distances(
            self.st,
            'SG',
            self.data_library.get_distances('SS_DIST'),
            not self.has_superimp_models()
        )

    def get_amide_list(self):
        """ Return list of amide residues """
        amide_res = self.data_library.get_amide_data()[0]

        amide_list = set()

        for res in self.st.get_residues():
            if res.get_resname() in amide_res:
                amide_list.add(res)

        return amide_list

    def get_chiral_list(self):
        """ Returns a list of chiral side chains"""
        chiral_res = self.data_library.get_chiral_data()
        chiral_list = []
        for res in self.st.get_residues():
            if res.get_resname() in chiral_res:
                chiral_list.append(res)
        return chiral_list

    def get_chiral_bck_list(self):
        """ Returns a list of residues with chiral CAs"""
        prot_chains = 0
        chiral_bck_list = []
        for chn in self.st.get_chains():
            if self.chain_ids[chn.id] == mu.PROTEIN:
                prot_chains += 1
                for res in chn.get_residues():
                    if res.get_resname() != 'GLY' and not mu.is_hetatm(res):
                        chiral_bck_list.append(res)
        if not prot_chains:
            print("No protein chains detected, skipping")
        return chiral_bck_list

    def check_r_list_clashes(self, residue_list, contact_types):
        """ Checks clashes originated by a list of residues"""
        return mu.check_r_list_clashes(
            residue_list,
            self.rr_dist,
            self.data_library.get_distances('CLASH_DIST'),
            self.data_library.get_atom_lists(contact_types),
            not self.has_superimp_models()
        )

    def check_rr_clashes(self, res1, res2, contact_types):
        """ Checks all clashes between two residues"""
        return mu.check_rr_clashes(
            res1,
            res2,
            self.data_library.get_distances('CLASH_DIST'),
            self.data_library.get_atom_lists(contact_types)
        )

    def check_missing_atoms(self):
        """ Makes a **list of missing atoms** in the structure

            Returns:
                List of residues with missing atoms, as a tuples
                ["r",{"backbone":[atoms_list],"side":[atoms_list]}]
        """
        #TODO Nucleic acids
        valid_codes = self.data_library.get_valid_codes('protein')
        residue_data = self.data_library.get_all_atom_lists()
        miss_at_list = []
        for res in self.st.get_residues():
            if res.get_resname() in valid_codes and not mu.is_hetatm(res):
                miss_at = mu.check_all_at_in_r(
                    res, residue_data[res.get_resname().replace(' ', '')]
                )
                if self.is_C_term(res) and 'OXT' not in res:
                    if 'backbone' not in miss_at:
                        miss_at['backbone'] = []
                    miss_at['backbone'].append('OXT')
                if miss_at:
                    miss_at_list.append([res, miss_at])
        return miss_at_list

    def get_missing_atoms(self, fragment):
        """ Returns list of residues with missing atoms

            Returns:
                List of residues with missing atoms, as a tuples like
                ["r",[atoms_list]]
        """
        miss_ats = []
        for res_at in self.check_missing_atoms():
            [res, at_list] = res_at
            if fragment == 'side':
                if 'side' in at_list and 'N' in res and 'CA' in res and 'C' in res:
                    miss_ats.append([res, at_list['side']])
            else:
                if at_list['backbone']:
                    miss_ats.append([res, at_list['backbone']])

        return miss_ats

    def get_ion_res_list(self):
        """
            returns list of residues with potencial selection on adding H

            Returns:
                List of residues that require selection on adding H
                ["r",[atom_list]]
        """
        ion_res = self.data_library.ion_res
        hydrogen_lists = self.data_library.get_hydrogen_atoms()

        ion_res_list = []
        for res in self.all_residues:
            rcode = res.get_resname()
            if rcode in ion_res:
                ion_res_list.append([res, hydrogen_lists[rcode]])
        return ion_res_list

    def get_backbone_breaks(self):
        """
            Determines several backbone anomalies
            1. Modified residues
            2. Backbone breaks
            3. Backbone links not corresponding to sequence
            4. Too long residue links
        """
        bck_breaks_list = []
        wrong_link_list = []
        not_link_seq_list = []
        self.modified_residue_list = []
        for i in range(0, len(self.all_residues)-1):
            res1 = self.all_residues[i]
            res2 = self.all_residues[i+1]
            if not mu.same_chain(res1, res2):
                continue
            if mu.is_hetatm(res1) or mu.is_hetatm(res2):
                if res1 in self.next_residue:
                    if self.next_residue[res1] == res2:
                        if mu.is_hetatm(res1):
                            self.modified_residue_list.append(res1)
                else:
                    continue
            #Skip NA #TODO include NA Backbone
            if self.chain_ids[res1.get_parent().id] != mu.PROTEIN:
                continue

            if res1 not in self.next_residue:
                bck_breaks_list.append([res1, res2])
                if mu.seq_consecutive(res1, res2):
                    dist = 0.
                    if 'N' in res1 and 'C' in res2:
                        dist = res1['N'] - res2['C']
                    not_link_seq_list.append([res1, res2, dist])

            else:
                if res2 != self.next_residue[res1]:
                    wrong_link_list.append([res1, self.next_residue[res1], res2])
        return {
            'bck_breaks_list': bck_breaks_list,
            'wrong_link_list': wrong_link_list,
            'not_link_seq_list': not_link_seq_list
        }

    def check_backbone_connect(self, backbone_atoms, covlnk):
        """
        Determines backbone links usign a distance criterium and produces a dict with
        link relationships in the N-term to C-term direction

        Args:
            backbone_atoms: atoms to be considered as backbone
            COVLNK: Threshold distance for a covalent bond
        """
        if not self.backbone_links:
            self.backbone_links = mu.get_backbone_links(
                self.get_structure(), backbone_atoms, covlnk
            )
        self.next_residue = {}
        self.prev_residue = {}
        for lnk in self.backbone_links:
            [at1, at2] = lnk
            res1 = at1.get_parent()
            res2 = at2.get_parent()
            self.prev_residue[res2] = res1
            self.next_residue[res1] = res2

    def check_cis_backbone(self):
        """
        Determines omega dihedrals for two bound residues and classifies them
        as normal trans, low angle trans, and cis

        Args:
            backbone_atoms: atoms to be considered as backbone
            covlnk: Distance threshold for a covalent bond

        """
        cis_backbone_list = []
        lowtrans_backbone_list = []
        for lnk in self.backbone_links:
            [at1, at2] = lnk
            res1 = at1.get_parent()
            res2 = at2.get_parent()
            if 'CA' in res1 and 'C' in res1 and 'CA' in res2 and 'N' in res2:
                dih = mu.calc_bond_dihedral(res1['CA'], res1['C'], res2['N'], res2['CA'])
                if abs(dih) < CISTHRES:
                    cis_backbone_list.append([res1, res2, dih])
                elif abs(dih) < TRANSTHRES:
                    lowtrans_backbone_list.append([res1, res2, dih])
        return (cis_backbone_list, lowtrans_backbone_list)

    def get_stats(self):
        """
         Returns a dict with calculates statistics

         Returns:
            Dict as {}
        """
        return {
            'nmodels': self.nmodels,
            'models_type': self.models_type,
            'nchains': len(self.chain_ids),
            'chain_ids': self.chain_ids,
            'num_res': self.num_res,
            'num_ats': self.num_ats,
            'res_insc': self.res_insc,
            'res_hetats': self.res_hetats,
            'res_ligands': self.res_ligands,
            'num_wat': self.num_wat,
            'ca_only': self.ca_only,
            'biounit': self.biounit
        }
    def print_headers(self):
        """
        Prints selected components from structure headers
        """
        self.get_headers()
        if 'entry_id'in self.meta:
            print(' PDB id: {}'.format(self.meta['entry_id']))
        print(' Title: {}'.format(self.meta['title']))
        print(' Experimental method: {}'.format(self.meta['method']))
        if 'keywords' in self.meta:
            print(' Keywords: {}'.format(self.meta['keywords']))
        if 'resolution' in self.meta:
            print(' Resolution: {} A'.format(self.meta['resolution']))
        if self.biounit:
            print(' Biounit no. {}'. format(self.meta['biounit']))

    def get_headers(self):
        """
        Extract selected components from structure headers
        """
        self.meta = {}
        if self.input_format == 'cif':
            self.meta['entry_id'] = self.headers['_entry.id']
            self.meta['title'] = self.headers['_struct.title']
            self.meta['method'] = self.headers['_exptl.method']
            self.meta['keywords'] = self.headers['_struct_keywords.pdbx_keywords']
            if '_refine_hist.d_res_high' in self.headers:
                self.meta['resolution'] = self.headers['_refine_hist.d_res_high']
        else:
            self.meta['title'] = self.headers['name']
            self.meta['method'] = self.headers['structure_method']
            if 'keywords' in self.headers:
                self.meta['keywords'] = self.headers['keywords']
            if 'resolution' in self.headers:
                self.meta['resolution'] = self.headers['resolution']
        if self.biounit:
            self.meta['biounit'] = self.biounit

    def print_stats(self, prefix=''):
        """
        Prints statistics to stdout

        Args:
            prefix: Text prefix to prepend to printed data
        """
        stats = self.get_stats()
        if stats['nmodels'] > 1:
            print(
                '{} Num. models: {} (type: {}, {:8.3f} A)'.format(
                    prefix,
                    stats['nmodels'],
                    mu.MODEL_TYPE_LABELS[stats['models_type']['type']],
                    stats['models_type']['rmsd']
                )
            )
        else:
            print('{} Num. models: {}'.format(prefix, stats['nmodels']))
        chids = []
        for ch_id in sorted(stats['chain_ids']):
            if isinstance(stats['chain_ids'][ch_id], list):
                chids.append('{}: Unknown '.format(ch_id))
            else:
                chids.append(
                    '{}: {}'.format(
                        ch_id, mu.CHAIN_TYPE_LABELS[stats['chain_ids'][ch_id]]
                    )
                )
        print('{} Num. chains: {} ({})'.format(prefix, stats['nchains'], ', '.join(chids)))
        print('{} Num. residues:  {}'.format(prefix, stats['num_res']))
        print('{} Num. residues with ins. codes:  {}'.format(prefix, stats['res_insc']))
        print('{} Num. HETATM residues:  {}'.format(prefix, stats['res_hetats']))
        print('{} Num. ligands or modified residues:  {}'.format(prefix, stats['res_ligands']))
        print('{} Num. water mol.:  {}'.format(prefix, stats['num_wat']))
        print('{} Num. atoms:  {}'.format(prefix, stats['num_ats']))
        if stats['ca_only']:
            print('Possible CA-Only structure')
        if self.hetatm[mu.MODRES]:
            print('Modified residues found')
            for res in self.hetatm[mu.MODRES]:
                print(mu.residue_id(res))
        if self.hetatm[mu.METAL]:
            print('Metal/Ion residues found')
            for res in self.hetatm[mu.METAL]:
                print(mu.residue_id(res))
        if self.hetatm[mu.ORGANIC]:
            print('Small mol ligands found')
            for res in self.hetatm[mu.ORGANIC]:
                print(mu.residue_id(res))


    def get_structure(self):
        """
        Method pointing to the enclosed Bio.PDB.Structure object

        Returns:
            a Bio.PDB.Structure object
        """
        return self.st

    def save_structure(self, output_pdb_path):
        """
        Saves structure on disk in PDB format

        Args:
            output_pdb_path: OS path to the output file

        Errors:
            OSError: Error saving the file
        """
        if not output_pdb_path:
            raise OutputPathNotProvidedError
        pdbio = PDBIO()
        pdbio.set_structure(self.st)
        pdbio.save(output_pdb_path)

# Methods to modify structure
    def select_model(self, keep_model):
        """ Selects one model and delete the others from the structure. Model is
            renumbered to model 1

            Args:
                keep_model: Model number to keep
        """
        ids = []
        for nmodel in self.st.get_models():
            ids.append(nmodel.id)
        for ind, model_id in enumerate(ids):
            if ind != keep_model - 1:
                self.st.detach_child(model_id)
        if keep_model != 1:
            self.st[keep_model-1].id = 0
        self.nmodels = 1
        self.models_type = 0
        # Update internal data
        self.update_internals()
        self.modified = True

    def has_models(self):
        """ Shotcut method to check whether the structure has more than one model

            Returns: Boolean
        """
        return self.nmodels > 1

    def has_superimp_models(self):
        """ Shotcut method to check whether the structure has superimposed
            models (i.e. NMR or ensemble)

            Returns: Boolean
        """
        return self.models_type and self.models_type['type'] == mu.ENSM

    def is_biounit(self):
        """Shortcut to check whether the structure has been loaded from a biounit
        """

    def set_chain_ids(self):
        """
        Identifies and sets the chain ids, guessing its nature (protein, dna, rna, ...)
        """
        self.chain_ids = {}
        for chn in self.st.get_chains():
            if not self.biounit and chn.get_parent().id > 0:
                continue
            self.chain_ids[chn.id] = mu.guess_chain_type(chn)
        self.sorted_chain_ids = sorted (self.chain_ids)

    def select_chains(self, select_chains):
        """
        Select one or more chains and remove the remaining.
        Args:
            select_chains: Comma separated chain ids
        """
        if not self.chain_ids:
            self.set_chain_ids()
        ch_ok = select_chains.split(',')
        for chn in ch_ok:
            if not chn in self.chain_ids:
                print('Warning: skipping unknown chain', chn)
        for mod in self.st:
            for chn in self.chain_ids:
                if chn not in ch_ok:
                    self.st[mod.id].detach_child(chn)
        # Update internal data
        self.update_internals()
        self.modified = True

    def select_altloc_residue(self, res, to_fix):
        """ Selects one alternative conformation when altloc exists. Selection is
            done on the occupancy basis or from the conformation id.
            All relevant atoms in the residue **res** at modified in the same way.
            Triggers **modified** flag.

            Args:
                res: residue (as Bio.PDB.Residue)
                to_fix: atoms in residue to fix as
                    {'ats':[atom_list],'select':'occupancy|conf_id'
        """
        for atm in to_fix['ats']:
            if to_fix['select'].lower() == 'occupancy':
                newat = atm.selected_child
            else:
                if to_fix['select'] in atm.child_dict.keys():
                    newat = atm.child_dict[to_fix['select']]
                else:
                    print(
                        'Warning: unknown alternative {} in {}'.format(
                            to_fix['select'], mu.atom_id(atm)
                        )
                    )
                    continue
            newat.disordered_flag = 0
            newat.altloc = ' '
            res.detach_child(atm.id)
            res.add(newat)
        res.disordered = 0
        self.modified = True

    def remove_residue(self, res, update_int=True):
        """
        Removes residue **r** from the structure. Triggers **modified** flag
        and atom and residue renumbering
        """
        mu.remove_residue(res)

        # Update internal data
        if update_int:
            self.update_internals()

        self.modified = True

    def fix_side_chain(self, r_at):
        """
        Fix missing side chain atoms in given residue. Triggers **modified** flag

        Args:
            **r_at**: tuple as [Bio.PDB.Residue, [list of atom ids]]
        """
        print(mu.residue_id(r_at[0]))
        for at_id in r_at[1]:
            print("  Adding new atom " + at_id)
            if at_id == 'CB':
                coords = mu.build_coords_CB(r_at[0])
            else:
                coords = mu.build_coords_from_lib(
                    r_at[0],
                    self.res_library,
                    r_at[0].get_resname(),
                    at_id
                )
            mu.add_new_atom_to_residue(r_at[0], at_id, coords)
        self.atom_renumbering()
        self.modified = True

    def fix_backbone_O_atoms(self, r_at):
        """Adding missing backbone atoms not affecting main-chain like O and OXT
                Args:
            **r_at**: tuple as [Bio.PDB.Residue, [list of atom ids]]
        """
        [res, at_list] = r_at
        print(mu.residue_id(res))
        if not 'C' in res:
            raise NotEnoughAtomsError
        if len(at_list) == 2 or at_list == ['O']:
            if 'CA' not in res or 'N' not in res or 'C' not in res:
                raise NotEnoughAtomsError
            print("  Adding new atom O")
            mu.add_new_atom_to_residue(res, 'O', mu.build_coords_O(res))
        if 'OXT' in at_list:
            if 'CA' not in res or 'C' not in res or 'O' not in res:
                raise NotEnoughAtomsError
            print("  Adding new atom OXT")
            mu.add_new_atom_to_residue(
                res,
                'OXT',
                mu.build_coords_SP2(mu.OINTERNALS[0], res['C'], res['CA'], res['O'])
            )

        self.atom_renumbering()
        self.modified = True
        return True

    def add_hydrogens(self, r_at_list, remove_h=True):
        """
        Add hydrogens according to selection in r_at_list

        Args:
           **r_at_list**: tuple as [Bio.PDB.Residue, Tauromeric Option]
           **hydrogens_list**: Hydrogen atom names per residue type
        """
        addH_rules = self.data_library.get_add_h_rules()

        done_side = set()

        if remove_h:
            for res in self.all_residues:
                mu.remove_H_from_r(res, verbose=False)

        # residues with alternative forms
        for r_at in r_at_list:
            [res, opt] = r_at

            if mu.is_hetatm(res):
                continue
        # Skip residues without addH rules
            if res.get_resname() not in addH_rules:
                print(NotAValidResidueError(res.get_resname()).message)
                done_side.add(res)
                continue
            rcode = res.get_resname()
            error_msg = mu.add_hydrogens_side(res, self.res_library, opt, addH_rules[rcode][opt])
            if error_msg:
                print(error_msg, mu.residue_id(res))
            res.resname = opt
            done_side.add(res)

        for res in self.all_residues:
            if mu.is_hetatm(res):
                continue

            rcode = res.get_resname()

            if res not in self.prev_residue:
                prev_residue = None
            else:
                prev_residue = self.prev_residue[res]

            error_msg = mu.add_hydrogens_backbone(res, prev_residue)
            if error_msg:
                print(error_msg, mu.residue_id(res))

            if res not in done_side and rcode != 'GLY':
                if rcode not in addH_rules:
                    print(NotAValidResidueError(rcode).message)
                    continue

                error_msg = mu.add_hydrogens_side(res, self.res_library, rcode, addH_rules[rcode])
                if error_msg:
                    print(error_msg, mu.residue_id(res))

        self.residue_renumbering()
        self.atom_renumbering()
        self.modified = True

    def is_N_term(self, res):
        """ Detects whether it is N terminal residue."""
        return res not in self.prev_residue

    def is_C_term(self, res):
        """ Detects whether it is C terminal residue."""
        return res not in self.next_residue

    def prepare_mutations(self, mut_list):
        """ Fins residues to mutate from mut_list"""
        mutations = MutationManager(mut_list)
        mutations.prepare_mutations(self.st)
        return mutations

    def apply_mutations(self, mutations):
        """ Perform mutations """
        mutated_res = mutations.apply_mutations(
            self.st,
            self.data_library.get_mutation_map(),
            self.res_library
        )
        self.residue_renumbering()
        self.atom_renumbering()
        self.modified = True
        return mutated_res

    def invert_amide_atoms(self, res):
        """ Fix sidechains with incorrect amide assignments"""
        amide_res = self.data_library.get_amide_data()[0]
        res_type = res.get_resname()
        if res_type not in amide_res:
            raise NotAValidResidueError(res_type)           
        mu.swap_atoms(
            res[amide_res[res_type][0]], 
            res[amide_res[res_type][1]]
        )
        
    def fix_chiral_chains(self, res):
        """ Fix sidechains with chiral errors"""
        chiral_res = self.data_library.get_chiral_data()
        res_type = res.get_resname()
        mu.swap_atoms(
            res[chiral_res[res_type][0]],
            res[chiral_res[res_type][1]]
        )
        if res_type == 'ILE':
            mu.delete_atom(res, 'CD1')
            mu.build_atom(res, 'CD1', self.res_library, 'ILE')

#===============================================================================

class Error(Exception):
    """ Base class """
    pass

class WrongServerError(Error):
    def __init__(self):
        self.message = 'ERROR: Biounits supported only on MMB server'

class UnknownFileTypeError(Error):
    def __init__(self, type):
        self.message = 'ERROR: unknown filetype ({})'.format(type)

class OutputPathNotProvidedError(Error):
    def __init__(self):
        self.message = 'ERROR: output PDB path not provided'

class NotAValidResidueError(Error):
    def __init__(self, res):
        self.message = 'Warning: {} is not a valid residue in this context'.format(res)
        
class NotEnoughAtomsError(Error):
    def __init__(self):
        self.message = 'Warning: not enough backbone to build missing atoms'



"""
    Class to support interaction with MMB PDB API to include support to biounits.
"""

import os
import sys

from Bio.PDB.PDBList import PDBList

from Bio._py3k import urlcleanup as _urlcleanup
from Bio._py3k import urlretrieve as _urlretrieve

URL_PREFIX = 'http://mmb.irbbarcelona.org/api/pdb'

class MMBPDBList(PDBList):
    """ Replacement class to support access to biounits at MMB PDB Server.
        
        Modified from original BioPython code.
    """

    def retrieve_pdb_file(
            self,
            pdb_code,
            obsolete=False,
            pdir=None,
            file_format=None,
            overwrite=False,
            biounit=False):
        """
            Replacement for Bio.PDB.PDBList.retrieve_pdb_file to support
            MMB PDB API. Defaults to Biopython super() if standard server is used.
        """
        if self.pdb_server != 'mmb':
            return super().retrieve_pdb_file(
                pdb_code, obsolete, pdir, file_format, overwrite
            )

        self._verbose = True

        code = pdb_code.lower()

        if file_format in ('pdb', 'mmCif', 'xml'):
            if file_format == 'mmCif':
                file_format = 'cif'
            if not biounit:
                url = (URL_PREFIX + '/%s.%s' % (code, file_format))
            else:
                file_format = 'pdb'
                url = (URL_PREFIX + '/%s_bn%s.pdb' % (code, biounit))
        else:
            print('Error: MMB Server: File format', file_format, 'not supported')
            sys.exit(1)
        #Where does the final PDB file get saved?
        if pdir is None:
            path = self.local_pdb if not obsolete else self.obsolete_pdb
            if not self.flat_tree:  # Put in PDB-style directory tree
                path = os.path.join(path, code[1:3])
        else:  # Put in specified directory
            path = pdir
        if not os.access(path, os.F_OK):
            os.makedirs(path)
        if biounit:
            final = {
                'pdb': '%s_%s.pdb',
                'mmCif': '%s_%s.cif',
                'cif': '%s_%s.cif',
                'xml': '%s_%s.xml'
            }
            final_file = os.path.join(path, final[file_format] % (code, biounit))
        else:
            final = {
                'pdb': '%s.pdb',
                'mmCif': '%s.cif',
                'cif': '%s.cif',
                'xml': '%s.xml'
            }
            final_file = os.path.join(path, final[file_format] % code)
        # Skip download if the file already exists
        if not overwrite:
            if os.path.exists(final_file):
                if self._verbose:
                    print("Structure exists: '%s' " % final_file)
                return final_file

        # Retrieve the file
        if self._verbose:
            if biounit:
                print("Downloading PDB structure '%s.%s'..." % (pdb_code, biounit))
            else:
                print("Downloading PDB structure '%s'..." % pdb_code)
        try:
            _urlcleanup()
            _urlretrieve(url, final_file)
        except IOError:
            print("Desired structure doesn't exists")
        return final_file



"""
 Module to read AMBER residue lib files
"""

import re
import sys

class ResidueLib():
    """ Class to manage a residue library taken from AMBER prep files.
    """
    def __init__(self, library_path):
        try:
            lib_file_h = open(library_path, "r")

        except IOError:
            print("ERROR: unable to open residue library" + library_path, file=sys.stderr)
            sys.exit(1)

        line = lib_file_h.readline()
        line = lib_file_h.readline()

        self.residues = {}
        at_group = False
        ch_group = False
        im_group = False
        res = ResidueDef()
        for line in lib_file_h:
            line = line.replace("\n", "").replace("\r", "")
            if line == '':
                continue

            elif line == "DONE":
                self.residues[res.id] = res
                at_group = False
                ch_group = False
                im_group = False
                res = ResidueDef()

            elif line == 'CHARGE':
                at_group = False
                ch_group = True

            elif line == 'IMPROPER':
                ch_group = False
                im_group = True

            elif re.match(' (...)  INT', line):
                resid_str = re.match(' (...)  INT', line)
                res.id = resid_str.group(1)
                at_group = True

            elif at_group:
                data = line.split()
                if len(data) < 11:
                    continue
                atm = AtomDef(data)
                for i in range(0, 3):
                    if atm.link[i] > 0:
                        atm.link_ats[i] = res.ats[atm.link[i]].id
                res.ats.append(atm)

            elif ch_group:
                for chrg in line.split():
                    res.charges.append(chrg)

            elif im_group:
                res.improper.append(line)

    def get_atom_def(self, res_id, at_id):
        """ Gets an atom definition given residue and atom names."""
        resid_def = self.residues[res_id]
        i = 1
        while resid_def.ats[i].id != at_id and i < len(resid_def.ats):
            i = i + 1
        if resid_def.ats[i].id == at_id:
            return resid_def.ats[i]
        return None

class ResidueDef():
    """ Class to pack a residue definition."""
    def __init__(self):
        self.id = ''
        self.name = ''
        self.ats = ['']
        self.improper = []
        self.charges = []

class AtomDef():
    """ Class to pack an atom definition."""
    def __init__(self, data):
        self.id = data[1]
        self.type = data[2]
        self.branch = data[3]
        self.link = [int(data[4]), int(data[5]), int(data[6])]
        self.link_ats = ['', '', '']
        self.geom = [float(data[7]), float(data[8]), float(data[9])]
        self.coord = [[0., 0., 0.], [0., 0., 0.], [0., 0., 0.]]
        self.chrg = float(data[10])

    def __str__(self):
        return self.id
